// Flash.c - flash memory routines for the Rex 6000
//     
// Copyright 2001 Chris Harris
//
// This file is part of Adder.
//
// Adder is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.  
//
// Adder is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
// GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License 
// along with Adder; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#asm
;
;	Copy flash memory page
;
;	8/1/2001 C.Harris
;
; Caller is responsible for validating page numbers,
; and verifying destination page is currently EMPTY ($FFs)!!!
; NOTE: parameters are PHYSICAL page numbers, 0000-00FF, 
; not LOGICAL (Rex-firmware-ish) page numbers ( = PP - 4).

	XDEF	_CopyFlashPage

._CopyFlashPage
	pop	bc
	pop	de	; destination page number
	pop	hl	; source page number
	push	hl
	push	de
	push	bc

	push	bc
	push	de

	ld	d,e	; D reg = destination page number (LO BYTE)
	ld	e,l	; E reg = source page number (LO BYTE)

	in	a,($03)
	ld	c,a
	in	a,($04)
	ld	b,a
	push	bc	; Save current bank 2 page setting
	push	ix

;	di			; disable interrupts during the copy

	ld	hl,$a000	; bank 2 address offset
 
.floop ld	a,e	; Source page
	out	($03),a
	xor	a
	out	($04),a

	ld	b,a
	ld	ix,$da3b
	
; COPY 256 bytes to low RAM - this is silly, but had to do it to use "flash hi"
; routine in Rex firmware (below)
.fsave	ld	a,(hl)	; get byte from source page
	ld	(ix+0),a
	inc	ix
	inc	l
	djnz fsave	


	ld	a,d	; Destination page
	out	($03),a
	xor	a
	out	($04),a

	ld	a,d
	cp	$80			; sector >= 80?
	jr	nc,flashhi	; Yes, use different flash routine!


; "Flash low": write to flash memory sector in FIRST MB of flash memory
;
; First, set "fast flash mode"
	ld	a,$aa
	ld	($0aaa),a	; Rex firmware has aaaa!?
	ld	a,$55
	ld	($0555),a	; ditto: a555
	ld	a,$20
	ld	($0aaa),a	; ditto: aaaa

	ld	ix,$da3b
	ld	b,0

.fwrite	ld	c,(ix+0)
	ld	(hl),$a0	; fast write command
	ld	(hl),c		; write byte to flash ROM

; wait for completion:
; (NOTE: this will likely fail if target location did not contain $FF!)
.fwait	ld      a,(hl)
	; New test(s) - from flash write logic @ $3760
	cp	c			; same as written: we are done with this byte
	jr	z,flonext
	bit	5,a			; this doesnt make any sense!
	jr	z,fwait
;	ld	a,(hl)
;	cp	c			; final verify
;	jr	z,flonext
; FAILED - for now, just continue on to next byte
;	jr	fdone		; FAILED: this byte still does not match, so give up (instead of hanging)

.flonext	inc	ix
	inc	l
	djnz	fwrite	; repeat for 256 bytes

; end "fast mode"
	dec     l
	ld      a,$90
	ld      (hl),a
	xor     a		; flash datasheet says this should be $f0 instead!?
	ld      (hl),a
	inc		l
	jr	fnext


; "Flash high": write to flash memory sector in SECOND MB of flash memory:
; not clear why I had to resort to this:  to flash sectors in SECOND MB
; of flash memory, have to use code executing in FIRST MB of flash! (and vice versa)
.flashhi	push	de
	push	hl
	push	hl		; target address (offset)
	ld	bc,$0100	; # bytes = 256
	ld	de,$da3b	; source address (in low RAM)
	call	$3760
	pop	af
	pop	hl
	pop	de


.fnext inc	h
	ld	a,h
	cp	$c0		; until address pointer rolls over to $C000
	jr	nz,floop	 ; go get another 256 bytes from source page
	

; All done w/ 8KB flash
.fdone	pop	ix
	pop	bc		; restore previous bank 2 page setting
	ld	a,c
	out	($03),a
	ld	a,b
	out	($04),a

;	ei	; and finally, re-enable interrupts

	pop	de
	pop	bc
	ret



;
;	EraseSector - erases the specified flash sector (64KB)
;
;	cah 7/31/2001

	XDEF	_EraseSector


._EraseSector
	pop	bc
	pop	hl	; sector number
	push	hl
	push	bc

	push de
	
;	di

	in  a,($03)		; save current Bank 2 page address
	ld	e,a
	in  a,($04)
	ld	d,a
	
	ld	a,l			; set new Bank 2 page address
	out	($03),a
	xor	a
	out	($04),a

	ld	a,l
	cp	$80			// sector >= 80?
	jr	nc,erasehi	// Yes, use different erase routine!

; LOW (within first MB) flash sector erase routine:
	ld	a,$aa
	ld	($0aaa),a
	ld	a,$55
	ld	($0555),a
	ld	a,$80
	ld	($0aaa),a
	ld	a,$aa
	ld	($0aaa),a
	ld	a,$55
	ld	($0555),a
	ld	hl,$a000
	ld	a,$30		; finally, erase the sector:
	ld	(hl),a		; pg:A000 = sector address

; this post-sector-erase test is from the firmware update logic at 1a3e
.etest1 ld	a,(hl)	; see if first byte erased okay
	cp	$ff
	jr	nz,etest1	; wait until it is $ff  (and, obviously: HANG HERE if erase fails!)
;
;	ld	c,(hl)	
;	cp	c		; make sure it STAYS $ff!?
;
	ld	a,(hl)	; why not use same test as above?	
	cp	$ff		; make sure it STAYS $ff!?
	jr	nz,etest1	; wait until it does
	jr	edone


; Although my "direct" sector erase routine above works fine in first 1MB flash memory,
; it FAILS in second 1MB flash memory.  I do not understand it, but we need to use
; the low-memory routine to do it, EVEN THOUGH IT IS ESSENTIALLY IDENTICAL to
; my routine.
.erasehi	call	$3723	; sector erase routine in Rex firmware, used by DSFW for storing addins


.edone 	ld	a,e
	out	($03),a
	ld	a,d
	out	($04),a

;	ei

	pop	de

	ret


;
;	IsPageBlank - checks the specified flash page (8KB)
;                     to see if it is currently empty = all $FFs.
;
;	cah 7/31/2001

	XDEF	_IsPageBlank


._IsPageBlank
	pop	bc
	pop	hl	; page number
	push	hl
	push	bc

	push de
	
	in  a,($03)
	ld	e,a
	in  a,($04)
	ld	d,a
	
	ld	a,l
	out	($03),a
	ld	a,h
	out	($04),a

; MUCH faster: do a quick check of just 2 bytes: first & middle
	ld	l,$ff
	ld	a,($a000)
	cp	l
	jr	nz,pnonEmpty
	ld	a,($b000)
	cp	l
	jr	nz,pnonEmpty


;	ld hl,$a000
;
;.ploop ld	a,(hl)
;	cp	$ff
;	jr	nz,pnonEmpty
;	inc	l
;	jr	nz,ploop
;	inc	h
;	ld	a,h
;	cp	$c0
;	jr	nz,ploop
;
	ld	l,1	; entire page is blank ($ff): return TRUE
	jr	pdone

.pnonEmpty ld l,0	; page is non-blank

.pdone ld	a,e
	out	($03),a
	ld	a,d
	out	($04),a

	pop	de

	ld	h,0	; HL = return flag
	ret


;
;	IsSectorBlank - checks the specified flash sector
;                     to see if it is currently empty = all $FFs.
;
;	cah 7/25/2001

	XDEF	_IsSectorBlank


._IsSectorBlank
	pop	bc
	pop	hl	; sector number
	push	hl
	push	bc

	push bc
	push de
	
	in  a,($03)
	ld	e,a
	in  a,($04)
	ld	d,a
	
;	ld	a,l
;	out	($03),a
	ld	a,h
	out	($04),a

	ld	b,8		; default to 8 pages = 64KB sector
	ld	a,l
	ld	c,a		; save sector # = starting page #
	cp	8		; is sector >= 8?
	jr	nc,bloop	; Yes, we are all set: go

	ld	b,1		; No, sector is 0-7, only 1 8KB page

.bloop	ld hl,$a000
	ld	a,c
	out	($03),a

.bloop2 ld	a,(hl)
	cp	$ff
	jr	nz,bnonEmpty
	inc	l
	jr	nz,bloop2
	inc	h
	ld	a,h
	cp	$c0
	jr	nz,bloop2
;
	inc	c
	djnz	bloop		; repeat for next 8KB page

	ld	l,1	; entire sector is blank ($ff): return TRUE
	jr	bdone

.bnonEmpty ld l,0	; sector is non-blank

.bdone ld	a,e
	out	($03),a
	ld	a,d
	out	($04),a

	pop	de
	pop	bc

	ld	h,0	; HL = return flag
	ret


;
;	"far" firmware Call (JUMP, actually) for REX6000
;
;	7/29/2001 C.Harris
;
		XDEF	_FarJump

._FarJump
	pop	bc
	pop	de	; offset
	pop	hl	; logical page number
	push	hl
	push	de
	push	bc
	ex	de,hl
	ld	a,e
	call	$59ca
	ret



; FlashByte - program a single byte in a specified 
;             page and offset in Rex flash memory

	XDEF	_FlashByte

._FlashByte
	pop	bc
	pop	de	; E = destination page number; D = byte to write
	pop	hl	; destination offset
	push	hl
	push	de
	push	bc

;	push	bc

	ld	bc,$2000
	add	hl,bc	; make HL = a000-based

	in	a,($03)	; Save current bank 2 page
	ld	c,a
	in	a,($04)
	ld	b,a

	ld	a,e		; Set bank 2 = Destination page
	out	($03),a
	xor	a
	out	($04),a

; "Flash low": write to flash memory sector in FIRST MB of flash memory
;
; First, set "fast flash mode"
	ld	a,$aa
	ld	($0aaa),a	; Rex firmware has aaaa!?
	ld	a,$55
	ld	($0555),a	; ditto: a555
	ld	a,$20
	ld	($0aaa),a	; ditto: aaaa

	ld	(hl),$a0	; fast write command
	ld	(hl),d		; write byte to flash ROM

; wait for completion:
; (NOTE: this will likely fail if target location did not contain $FF!)
.fbwait	ld      a,(hl)
	; New test(s) - from flash write logic @ $3760
	cp	d			; same as written: we are done with this byte
	jr	z,fbgood
	bit	5,a			; this doesnt make any sense!
	jr	z,fbwait

.fbgood	ld  a,$90
	ld      (hl),a
	xor     a		; flash datasheet says this should be $f0 instead!?
	ld      (hl),a
		
	ld	a,c			; restore previous bank 2 page setting
	out	($03),a
	ld	a,b
	out	($04),a

;	pop	bc
	ret





; Peek - retrieve a single byte from a specified 
;        page and offset in Rex flash memory

	XDEF	_Peek

._Peek
	pop	bc	; return address
	pop	hl	; offset
	pop	de	; page number
	push	de
	push	hl
	push	bc

;	push	bc

	ld	bc,$2000
	add	hl,bc	; make HL = a000-based

	in	a,($03)	; Save current bank 2 page
	ld	c,a
	in	a,($04)
	ld	b,a

	ld	a,e		; Set bank 2 = Destination page
	out	($03),a
	xor	a
	out	($04),a

	ld	e,(hl)	; get two bytes
	inc	hl
	ld	d,(hl)
		
	ld	a,c			; restore previous bank 2 page setting
	out	($03),a
	ld	a,b
	out	($04),a

;	pop	bc
	ld	l,e			; return word in HL
	ld	h,d
	ret


#endasm

